import collections
import inspect
from abc import ABC
from contextlib import contextmanager
from types import MethodType
from typing import List, Callable, Iterable, Set, Dict, Type, Any, Union, Optional, NamedTuple

from mcdreforged.command.builder import command_builder_util as utils
from mcdreforged.command.builder.exception import LiteralNotMatch, UnknownCommand, UnknownArgument, CommandSyntaxError, \
	UnknownRootArgument, RequirementNotMet, IllegalNodeOperation, \
	CommandError
from mcdreforged.command.command_source import CommandSource
from mcdreforged.utils.types import MessageText

__SOURCE_CONTEXT_CALLBACK = Union[Callable[[], Any], Callable[[CommandSource], Any], Callable[[CommandSource, dict], Any]]
__SOURCE_CONTEXT_CALLBACK_BOOL = Union[Callable[[], bool], Callable[[CommandSource], bool], Callable[[CommandSource, dict], bool]]
__SOURCE_CONTEXT_CALLBACK_MSG = Union[Callable[[], MessageText], Callable[[CommandSource], MessageText], Callable[[CommandSource, dict], MessageText]]
__SOURCE_CONTEXT_CALLBACK_STR_ITERABLE = Union[Callable[[], Iterable[str]], Callable[[CommandSource], Iterable[str]], Callable[[CommandSource, dict], Iterable[str]]]
__SOURCE_ERROR_CONTEXT_CALLBACK = Union[Callable[[], Any], Callable[[CommandSource], Any], Callable[[CommandSource, CommandError], Any], Callable[[CommandSource, CommandError, dict], Any]]

RUNS_CALLBACK = __SOURCE_CONTEXT_CALLBACK
ERROR_HANDLER_CALLBACK = __SOURCE_ERROR_CONTEXT_CALLBACK
FAIL_MSG_CALLBACK = __SOURCE_CONTEXT_CALLBACK_MSG
SUGGESTS_CALLBACK = __SOURCE_CONTEXT_CALLBACK_STR_ITERABLE
REQUIRES_CALLBACK = __SOURCE_CONTEXT_CALLBACK_BOOL


class ParseResult(NamedTuple):
	value: Optional[Any]
	char_read: int


class CommandSuggestion:
	def __init__(self, command_read: str, suggest_segment: str):
		self.__command_read = command_read
		self.__suggest_segment = suggest_segment

	def __hash__(self):
		return hash(self.__suggest_segment) + hash(self.__command_read) * 31

	def __eq__(self, other):
		return isinstance(other, type(self)) and self.__dict__ == other.__dict__

	@property
	def command(self) -> str:
		return self.__command_read + self.__suggest_segment

	@property
	def existed_input(self) -> str:
		return self.__command_read

	@property
	def suggest_input(self) -> str:
		return self.__suggest_segment

	def __str__(self):
		return '{} -> {}'.format(self.__command_read, self.__suggest_segment)


class CommandSuggestions(List[CommandSuggestion]):
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		# !!MCDR plg load <file_name>
		#                 ^
		#               cursor
		# "<file_name>" is the complete_hint
		self.complete_hint: Optional[str] = None

	def extend(self, __iterable: Iterable) -> None:
		super().extend(__iterable)
		if isinstance(__iterable, CommandSuggestions):
			self.complete_hint = self.complete_hint or __iterable.complete_hint


class CommandContext(Dict[str, Any]):
	def __init__(self, source: CommandSource, command: str):
		super().__init__()
		self.__source = source
		self.__command = command
		self.__cursor = 0
		self.__node_path = []  # type: List[AbstractNode]

	@property
	def source(self) -> CommandSource:
		return self.__source

	@property
	def command(self) -> str:
		return self.__command

	@property
	def command_read(self) -> str:
		return self.__command[:self.__cursor]

	@property
	def command_remaining(self) -> str:
		return self.__command[self.__cursor:]

	@property
	def cursor(self) -> int:
		return self.__cursor

	@property
	def node_path(self) -> List['AbstractNode']:
		return self.__node_path

	@contextmanager
	def read_command(self, current_node: 'AbstractNode', result: ParseResult, new_cursor: int):
		"""
		**Only used in command parsing**
		Change the current cursor position, and store the parsing value
		"""
		prev_cursor = self.__cursor
		self.__cursor = new_cursor
		if isinstance(current_node, ArgumentNode):
			self[current_node.get_name()] = result.value
		try:
			yield
		finally:
			self.__cursor = prev_cursor
			if isinstance(current_node, ArgumentNode):
				self.pop(current_node.get_name(), None)

	@contextmanager
	def enter_child(self, node: 'AbstractNode'):
		"""
		**Only used in command parsing**
		Enter a command node, maintain the node_path
		"""
		self.__node_path.append(node)
		try:
			yield
		finally:
			self.__node_path.pop(len(self.__node_path) - 1)


class _ErrorHandler(NamedTuple):
	callback: ERROR_HANDLER_CALLBACK
	handled: bool


_ERROR_HANDLER_TYPE = Dict[Type[CommandError], _ErrorHandler]


class AbstractNode(ABC):

	def __init__(self):
		self._children_literal: Dict[str, List[Literal]] = collections.defaultdict(list)  # mapping from literal text to related Literal nodes
		self._children: List[AbstractNode] = []
		self._callback: Optional[RUNS_CALLBACK] = None
		self._error_handlers: _ERROR_HANDLER_TYPE = {}
		self._child_error_handlers: _ERROR_HANDLER_TYPE = {}
		self._requirement: REQUIRES_CALLBACK = lambda source: True
		self._requirement_failure_message_getter: Optional[FAIL_MSG_CALLBACK] = None
		self._redirect_node: Optional[AbstractNode] = None
		self._suggestion_getter: SUGGESTS_CALLBACK = lambda: []

	# --------------
	#   Interfaces
	# --------------

	def then(self, node: 'AbstractNode') -> 'AbstractNode':
		"""
		Add a child node to this node
		:param node: a child node for new level command
		"""
		if self._redirect_node is not None:
			raise IllegalNodeOperation('Redirected node is not allowed to add child nodes')
		if isinstance(node, Literal):
			for literal in node.literals:
				self._children_literal[literal].append(node)
		else:
			self._children.append(node)
		return self

	def runs(self, func: RUNS_CALLBACK) -> 'AbstractNode':
		"""
		Executes the given function if the command string ends here
		:param func: A function to execute at this node which accepts maximum 2 parameters (command source and context)
		:rtype: AbstractNode
		"""
		self._callback = func
		return self

	def requires(self, requirement: REQUIRES_CALLBACK, failure_message_getter: Optional[FAIL_MSG_CALLBACK] = None) -> 'AbstractNode':
		"""
		Set the requirement for the command source to enter this node
		:param requirement: A callable function which accepts maximum 2 parameters (command source and context)
		and return a bool indicating whether the source is allowed to executes this command or not
		:param failure_message_getter: The reason to show in the exception when the requirement is not met.
		It's a callable function which accepts maximum 2 parameters (command source and context). If it's not specified,
		a default message will be used
		:rtype: AbstractNode
		"""
		self._requirement = requirement
		self._requirement_failure_message_getter = failure_message_getter
		return self

	def redirects(self, redirect_node: 'AbstractNode') -> 'AbstractNode':
		"""
		Redirect the child branches of this node to the child branches of the given node
		:type redirect_node: AbstractNode
		:rtype: AbstractNode
		"""
		if self.has_children():
			raise IllegalNodeOperation('Node with children nodes is not allowed to be redirected')
		self._redirect_node = redirect_node
		return self

	def suggests(self, suggestion: SUGGESTS_CALLBACK) -> 'AbstractNode':
		"""
		Set the provider for command suggestions of this node
		:param suggestion: A callable function which accepts maximum 2 parameters (command source and context)
		and return an iterable of str indicating the current command suggestions
		:rtype: AbstractNode
		"""
		self._suggestion_getter = suggestion
		return self

	def on_error(self, error_type: Type[CommandError], handler: ERROR_HANDLER_CALLBACK, *, handled: bool = False) -> 'AbstractNode':
		"""
		When a command error occurs, invoke the handler
		:param error_type: A class that is subclass of CommandError
		:param handler: A callback function which accepts maximum 3 parameters (command source, error and context)
		:param handled: If handled is set to True, error.set_handled() is called automatically when invoking the handler callback
		"""
		if not issubclass(error_type, CommandError):
			raise TypeError('error_type parameter should be a class inherited from CommandError, but class {} found'.format(error_type))
		self._error_handlers[error_type] = _ErrorHandler(handler, handled)
		return self

	def on_child_error(self, error_type: Type[CommandError], handler: ERROR_HANDLER_CALLBACK, *, handled: bool = False) -> 'AbstractNode':
		"""
		When receives a command error from one of the node's child, invoke the handler
		:param error_type: A class that is subclass of CommandError
		:param handler: A callback function which accepts maximum 3 parameters (command source, error and context)
		:param handled: If handled is set to True, error.set_handled() is called automatically when invoking the handler callback
		"""
		if not issubclass(error_type, CommandError):
			raise TypeError('error_type parameter should be a class inherited from CommandError, but class {} found'.format(error_type))
		self._child_error_handlers[error_type] = _ErrorHandler(handler, handled)
		return self

	# -------------------
	#   Interfaces ends
	# -------------------

	def _get_usage(self) -> str:
		raise NotImplementedError()

	def has_children(self):
		return len(self._children) + len(self._children_literal) > 0

	def parse(self, text: str) -> ParseResult:
		"""
		Try to parse the text and get a argument. Return a ParseResult instance indicating the parsing result
		ParseResult.value: The value to store in the context dict
		ParseResult.remaining: The remain
		:param str text: the remaining text to be parsed. It's supposed to not be started with DIVIDER character
		"""
		raise NotImplementedError()

	@staticmethod
	def __smart_callback(callback: Callable, *args):
		sig = inspect.signature(callback)
		spec_args = inspect.getfullargspec(callback).args
		spec_args_len = len(spec_args)
		if isinstance(callback, MethodType):  # class method, remove the 1st param
			spec_args_len -= 1
		try:
			sig.bind(*args[:spec_args_len])  # test if using full arg length is ok
		except TypeError:
			raise
		return callback(*args[:spec_args_len])

	def __handle_error(self, error: CommandError, context: CommandContext, error_handlers: _ERROR_HANDLER_TYPE):
		for error_type, handler in error_handlers.items():
			if isinstance(error, error_type):
				self.__smart_callback(handler.callback, context.source, error, context)
				if handler.handled:
					error.set_handled()

	def __raise_error(self, error: CommandError, context: CommandContext):
		self.__handle_error(error, context, self._error_handlers)
		raise error

	def _get_suggestions(self, context: CommandContext) -> Iterable[str]:
		return self.__smart_callback(self._suggestion_getter, context.source, context)

	def _execute_command(self, context: CommandContext) -> None:
		command = context.command  # type: str
		try:
			parse_result = self.parse(context.command_remaining)
		except CommandSyntaxError as error:
			error.set_parsed_command(context.command_read)
			error.set_failed_command(context.command_read + context.command_remaining[:error.char_read])
			self.__raise_error(error, context)
		else:
			next_remaining = utils.remove_divider_prefix(context.command_remaining[parse_result.char_read:])  # type: str
			total_read = len(command) - len(next_remaining)  # type: int

			with context.read_command(self, parse_result, total_read):
				if self._requirement is not None:
					if not self.__smart_callback(self._requirement, context.source, context):
						getter = self._requirement_failure_message_getter or (lambda: None)
						failure_message = self.__smart_callback(getter, context.source, context)
						self.__raise_error(RequirementNotMet(context.command_read, context.command_read, failure_message), context)

				# Parsing finished
				if len(next_remaining) == 0:
					if self._callback is not None:
						self.__smart_callback(self._callback, context.source, context)
					else:
						self.__raise_error(UnknownCommand(context.command_read, context.command_read), context)
				# Un-parsed command string remains
				else:
					# Redirecting
					node = self if self._redirect_node is None else self._redirect_node

					argument_unknown = False
					# No child at all
					if not node.has_children():
						argument_unknown = True
					else:
						# Pass the remaining command string to the children
						next_literal = utils.get_element(next_remaining)
						try:
							# Check literal children first
							literal_error = None
							for child_literal in node._children_literal.get(next_literal, []):
								try:
									with context.enter_child(child_literal):
										child_literal._execute_command(context)
									break
								except CommandError as e:
									# it's ok for a direct literal node to fail
									# other literal might still have a chance to consume this command
									literal_error = e
							else:  # All literal children fails
								if literal_error is not None:
									raise literal_error
								for child in node._children:
									with context.enter_child(child):
										child._execute_command(context)
									break
								else:  # No argument child
									argument_unknown = True

						except CommandError as error:
							self.__handle_error(error, context, self._child_error_handlers)
							raise error from None

					if argument_unknown:
						self.__raise_error(UnknownArgument(context.command_read, command), context)

	def _generate_suggestions(self, context: CommandContext) -> CommandSuggestions:
		"""
		Return a list of tuple (suggested command, suggested argument)
		"""
		def self_suggestions():
			return CommandSuggestions([CommandSuggestion(command_read_at_the_beginning, s) for s in self._get_suggestions(context)])

		suggestions = CommandSuggestions()
		# [!!aa bb cc] dd
		# read         suggested
		command_read_at_the_beginning = context.command_read
		if len(context.command_remaining) == 0:
			return self_suggestions()
		try:
			result = self.parse(context.command_remaining)
		except CommandSyntaxError:
			return self_suggestions()
		else:
			success_read = len(context.command) - len(context.command_remaining) + result.char_read  # type: int
			next_remaining = utils.remove_divider_prefix(context.command_remaining[result.char_read:])  # type: str
			total_read = len(context.command) - len(next_remaining)  # type: int

			with context.read_command(self, result, total_read):
				if self._requirement is not None and not self.__smart_callback(self._requirement, context.source, context):
					return CommandSuggestions()

				# Parsing finished
				if len(next_remaining) == 0:
					# total_read == success_read means DIVIDER does not exists at the end of the input string
					# in that case, ends at this current node
					if success_read == total_read:
						return self_suggestions()

				node = self if self._redirect_node is None else self._redirect_node
				# Check literal children first
				children_literal = node._children_literal.get(utils.get_element(next_remaining), [])
				for child_literal in children_literal:
					with context.enter_child(child_literal):
						suggestions.extend(child_literal._generate_suggestions(context))
				if len(children_literal) == 0:
					for literal_list in node._children_literal.values():
						for child_literal in literal_list:
							with context.enter_child(child_literal):
								suggestions.extend(child_literal._generate_suggestions(context))
					usages = []
					for child in node._children:
						with context.enter_child(child):
							suggestions.extend(child._generate_suggestions(context))
							if len(next_remaining) == 0:
								usages.append(child._get_usage())
					if len(next_remaining) == 0:
						suggestions.complete_hint = '|'.join(usages)
		return suggestions


class EntryNode(AbstractNode, ABC):
	def execute(self, source, command):
		"""
		Parse and execute this command
		Raise variable CommandError if parsing fails
		:param CommandSource source: the source that executes this command
		:param str command: the command string to execute
		"""
		try:
			self._execute_command(CommandContext(source, command))
		except LiteralNotMatch as error:
			# the root literal node fails to parse the first element
			raise UnknownRootArgument(error.get_parsed_command(), error.get_failed_command()) from error

	def generate_suggestions(self, source, command) -> CommandSuggestions:
		"""
		Get a list of command suggestion of given command
		Return an empty list if parsing fails
		:param CommandSource source: the source that executes this command
		:param str command: the command string to execute
		"""
		return self._generate_suggestions(CommandContext(source, command))


class Literal(EntryNode):
	"""
	A literal argument, doesn't store any value, only for extending and readability of the command
	The only argument type that is allowed to use the execute method
	"""
	def __init__(self, literal: str or Iterable[str]):
		super().__init__()
		if isinstance(literal, str):
			literals = {literal}
		elif isinstance(literal, Iterable):
			literals = set(literal)
		else:
			raise TypeError('Only str or Iterable[str] is accepted')
		for literal in literals:
			if not isinstance(literal, str):
				raise TypeError('Literal node only accepts str but {} found'.format(type(literal)))
			if utils.DIVIDER in literal:
				raise TypeError('DIVIDER character "{}" cannot be inside a literal'.format(utils.DIVIDER))
		self.literals = literals  # type: Set[str]
		self._suggestion_getter = lambda: self.literals

	def _get_usage(self) -> str:
		return '|'.join(sorted(self.literals))

	def suggests(self, suggestion: SUGGESTS_CALLBACK) -> 'AbstractNode':
		raise IllegalNodeOperation('Literal node doe not support suggests')

	def parse(self, text):
		arg = utils.get_element(text)
		if arg in self.literals:
			return ParseResult(None, len(arg))
		else:
			raise LiteralNotMatch('Invalid Argument', len(arg))

	def __repr__(self):
		return 'Literal[literals={}]'.format(self.literals)


class ArgumentNode(AbstractNode, ABC):
	def __init__(self, name: str):
		super().__init__()
		self.__name = name

	def get_name(self) -> str:
		return self.__name

	def _get_usage(self) -> str:
		return '<{}>'.format(self.__name)



import os

from setuptools import find_packages, setup

from mcdreforged.constants import core_constant

# === cheatsheet ===
# rm -rf build/ dist/ mcdreforged.egg-info/
# python setup.py sdist bdist_wheel
# python -m twine upload --repository testpypi dist/*
# python -m twine upload dist/*

NAME = core_constant.PACKAGE_NAME
VERSION = core_constant.VERSION_PYPI
DESCRIPTION = 'A rewritten version of MCDaemon, a python script to control your Minecraft server'
URL = 'https://github.com/Fallen-Breath/MCDReforged'
AUTHOR = 'Fallen_Breath'
REQUIRES_PYTHON = '>=3.6.0'

CLASSIFIERS = [
	# https://pypi.org/classifiers/
	'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
	'Programming Language :: Python',
	'Programming Language :: Python :: 3',
	'Programming Language :: Python :: 3.6',
	'Programming Language :: Python :: 3.7',
	'Programming Language :: Python :: 3.8',
	'Programming Language :: Python :: 3.9',
	'Operating System :: OS Independent'
]

if os.getenv('CI', None) is not None:
	build_num = os.getenv('GITHUB_RUN_NUMBER', None)
	is_release = os.getenv('GITHUB_REF', '').startswith('refs/tags/v')
	if build_num is not None and not is_release:
		VERSION += '.dev{}'.format(build_num)

# ----------------------------------------------------------------

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'requirements.txt')) as f:
	REQUIRED = [line for line in f.readlines() if not len(line.strip()) == 0]

print('REQUIRED = {}'.format(REQUIRED))

with open(os.path.join(here, 'README.md'), encoding='utf8') as f:
	LONG_DESCRIPTION = f.read()


setup(
	name=NAME,
	version=VERSION,
	description=DESCRIPTION,
	long_description=LONG_DESCRIPTION,
	long_description_content_type='text/markdown',
	author=AUTHOR,
	python_requires=REQUIRES_PYTHON,
	url=URL,
	packages=find_packages(exclude=['tests', '*.tests', '*.tests.*', 'tests.*']),
	install_requires=REQUIRED,
	include_package_data=True,
	classifiers=CLASSIFIERS,
)



from typing import TYPE_CHECKING, NamedTuple, Callable, Collection, Any, Type, Dict, List, Optional

from mcdreforged.command.builder.exception import RequirementNotMet
from mcdreforged.command.builder.nodes.arguments import QuotableText
from mcdreforged.command.builder.nodes.basic import Literal, ArgumentNode
from mcdreforged.command.command_source import CommandSource
from mcdreforged.minecraft.rtext import RText, RColor, RAction, RStyle, RTextList, RTextBase
from mcdreforged.plugin.builtin.mcdreforged_plugin.commands.sub_command import SubCommand
from mcdreforged.preference.preference_manager import PreferenceItem, PreferenceManager

if TYPE_CHECKING:
	from mcdreforged.plugin.builtin.mcdreforged_plugin.mcdreforged_plugin import MCDReforgedPlugin


PREF_COLOR = RColor.dark_aqua
VALUE_COLOR = RColor.gold


class PrefCommandEntry(NamedTuple):
	pref_name: str
	node_type: Type[ArgumentNode]
	suggester: Callable[[], Collection[str]]
	callback: Callable[[CommandSource, Any], Any]


class PreferenceCommand(SubCommand):
	def __init__(self, mcdr_plugin: 'MCDReforgedPlugin'):
		super().__init__(mcdr_plugin)

		def add_preference(entry: PrefCommandEntry):
			self.preferences[entry.pref_name] = entry

		self.preferences: Dict[str, PrefCommandEntry] = {}
		add_preference(PrefCommandEntry('language', QuotableText, lambda: self.available_languages, self.set_language))

	@property
	def available_languages(self) -> Collection[str]:
		return self.mcdr_server.translation_manager.available_languages

	@property
	def pref_mgr(self) -> 'PreferenceManager':
		return self.mcdr_server.preference_manager

	def allowed_to_use(self, source: CommandSource) -> bool:
		try:
			self.pref_mgr.get_preference(source, strict_type_check=True)
		except TypeError:
			return False
		else:
			return True

	def get_command_node(self) -> Literal:
		root_node = (
			self.public_command_root({'preference', 'pref'}).
			runs(lambda src: src.reply(self.get_help_message(src, 'mcdr_command.help_message.preference'))).
			requires(lambda src: self.allowed_to_use).
			on_error(RequirementNotMet, lambda src: src.reply(self.tr('mcdr_command.preference.unsupported_command_source', type(src)).set_color(RColor.red)), handled=True).
			then(Literal('list').runs(lambda src: self.show_preference_list(src)))
		)
		for pref in self.preferences.values():
			root_node.then(
				Literal(pref.pref_name).
				runs(lambda src, ctx: self.show_preference_item(src, pref.pref_name)).
				then(Literal('set').then(
					pref.node_type('value').
					suggests(pref.suggester).
					runs(lambda src, ctx: pref.callback(src, ctx['value']))
				)).
				then(Literal('reset').runs(lambda src: pref.callback(src, None)))
			)
		return root_node

	def __detail_hint(self, text: RTextBase, pref_name: str) -> RTextBase:
		return (
			text.
			h(self.tr('mcdr_command.preference.list.detail_hint', RText(pref_name, PREF_COLOR))).
			c(RAction.run_command, '{} preference {}'.format(self.control_command_prefix, pref_name))
		)

	def show_preference_list(self, source: CommandSource):
		pref = self.pref_mgr.get_preference(source, auto_add=True)
		source.reply(self.tr('mcdr_command.preference.list.title'))
		for pref_name in pref.get_annotations_fields().keys():
			value = getattr(pref, pref_name, RText('N/A', RColor.gray))
			source.reply(self.__detail_hint(
				RTextList(RText('- ', RColor.gray), RText(pref_name, PREF_COLOR), RText(': ', RColor.gray), RText(value, VALUE_COLOR)),
				pref_name
			))

	def show_preference_item(self, source: CommandSource, pref_name: str):
		entry: PrefCommandEntry = self.preferences.get(pref_name)
		pref: PreferenceItem = self.pref_mgr.get_preference(source, auto_add=True)
		current_value = getattr(pref, pref_name, None)
		default_value = getattr(self.pref_mgr.get_default_preference(), pref_name, None)
		if entry is None or current_value is None:  # should never come here
			raise KeyError('Unknown preference {}'.format(pref_name))

		def get_suggestion_text(value: str):
			text = RText(value, VALUE_COLOR).c(RAction.suggest_command, '{} preference {} set {}'.format(self.control_command_prefix, pref_name, value))
			hover_text = RTextList(self.tr('mcdr_command.preference.item.set_suggestion_hint', RText(pref_name, PREF_COLOR), RText(value, VALUE_COLOR)))
			styles: List[RStyle] = []
			extra_descriptions: List[RTextBase] = []
			if value == current_value:
				styles.append(RStyle.underlined)
				extra_descriptions.append(self.tr('mcdr_command.preference.item.current').set_color(RColor.gray))
			if value == default_value:
				styles.append(RStyle.bold)
				extra_descriptions.append(self.tr('mcdr_command.preference.item.default').set_color(RColor.gray))
			if len(extra_descriptions) > 0:
				hover_text.append(RTextList('\n', RText.join(RText(', ', RColor.dark_gray), extra_descriptions)))
			text.set_styles(styles)
			return text.h(hover_text)

		source.reply(self.tr('mcdr_command.preference.item.name', self.__detail_hint(RText(pref_name, PREF_COLOR), pref_name)))
		source.reply(self.tr('mcdr_command.preference.item.value', RText(current_value, VALUE_COLOR)))
		source.reply(self.tr('mcdr_command.preference.item.suggestions', RText.join(', ', map(get_suggestion_text, entry.suggester()))))

	def set_language(self, source: CommandSource, new_lang: Optional[str]):
		if new_lang is None:  # reset
			new_lang = self.pref_mgr.get_default_preference().language
		if new_lang not in self.available_languages:
			source.reply(self.tr('mcdr_command.preference.unknown_language', new_lang))
			return
		pref = self.pref_mgr.get_preference(source, auto_add=True)
		pref.language = new_lang
		self.pref_mgr.save_preferences()
		source.reply(self.tr('mcdr_command.preference.set.done', RText('language', RColor.yellow), RText(new_lang, RColor.gold)))




"""
Constants related to plugin for MCDR
"""

# the directory inside MCDR working directory that stores plugins' configure files
PLUGIN_CONFIG_DIRECTORY = 'config'

# The file prefix for a solo plugin (a single .py file)
SOLO_PLUGIN_FILE_SUFFIX = '.py'

# The file prefix for a packed plugin (a zipped file)
PACKED_PLUGIN_FILE_SUFFIXES = ('.mcdr', '.pyz')

# The name of the meta file inside a packed plugin
PLUGIN_META_FILE = 'mcdreforged.plugin.json'

# The name of the python package requirement file
PLUGIN_REQUIREMENTS_FILE = 'requirements.txt'

# the path to the directory inside a packed plugin as the default directory to store the language files of the plugin
PLUGIN_TRANSLATION_FILES_PATH = 'lang'

# Disabled plugin file name suffix
DISABLED_PLUGIN_FILE_SUFFIX = '.disabled'



import importlib.machinery
import importlib.machinery
import importlib.util
import os
import re
import sys
from typing import TYPE_CHECKING

from mcdreforged.constants import plugin_constant
from mcdreforged.plugin.meta.metadata import Metadata
from mcdreforged.plugin.type.regular_plugin import RegularPlugin
from mcdreforged.utils import string_util

if TYPE_CHECKING:
	from mcdreforged.plugin.plugin_manager import PluginManager


class SoloPlugin(RegularPlugin):
	def __init__(self, plugin_manager: 'PluginManager', file_path: str):
		super().__init__(plugin_manager, file_path)
		self.module_name = '__MCDR_SOLO_PLUGIN__@{}'.format(self.plugin_path)

	def get_fallback_metadata_id(self) -> str:
		file_name = string_util.remove_suffix(self.file_name, plugin_constant.SOLO_PLUGIN_FILE_SUFFIX)
		return re.sub(r'[^a-z0-9]', '_', file_name.lower())

	def is_own_module(self, module_name: str) -> bool:
		return module_name == self.module_name

	def _get_module_instance(self):
		if not os.path.isfile(self.plugin_path):
			raise TypeError('Source path {} of {} is not a file'.format(self.plugin_path, self))

		# https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
		# https://docs.python.org/zh-cn/3.6/library/importlib.html#importing-a-source-file-directly
		spec = importlib.util.spec_from_file_location(self.module_name, self.plugin_path)
		module = importlib.util.module_from_spec(spec)
		# noinspection PyUnresolvedReferences
		spec.loader.exec_module(module)
		# store the module, in case something want to access sys.modules[some_field.__module__]
		# e.g. typing.get_type_hints
		sys.modules[self.module_name] = module
		return module

	def _on_load(self):
		super()._on_load()
		self._load_entry_instance()
		self._set_metadata(Metadata(getattr(self.entry_module_instance, 'PLUGIN_METADATA', None), plugin=self))



from parse import parse

from mcdreforged.handler.impl.abstract_minecraft_handler import AbstractMinecraftHandler


class BukkitHandler(AbstractMinecraftHandler):
	@classmethod
	def get_content_parsing_formatter(cls):
		return '[{hour:d}:{min:d}:{sec:d} {logging}]: {content}'

	def parse_server_stdout(self, text):
		result = super().parse_server_stdout(text)
		if result.player is None:
			parsed = parse('[{dim_name}]<{name}> {message}', result.content)
			if parsed is not None and self._verify_player_name(parsed['name']):
				result.player, result.content = parsed['name'], parsed['message']
		return result




import abc
from typing import Dict, List, Callable, TYPE_CHECKING

if TYPE_CHECKING:
	from mcdreforged.plugin.type.plugin import AbstractPlugin


class PluginEvent(abc.ABC):
	def __init__(self, event_id: str):
		self.id = event_id


class LiteralEvent(PluginEvent):
	def __init__(self, event_id: str):
		super().__init__(event_id)

	def __repr__(self):
		return 'LiteralEvent[id={}]'.format(self.id)


class MCDREvent(PluginEvent):
	def __init__(self, event_id: str, name: str, default_method_name: str):
		super().__init__(event_id)
		self.name = name
		self.default_method_name = default_method_name

	def __repr__(self):
		return 'MCDREvent[name={}]'.format(self.name)


class _PluginEventStorage:
	EVENT_DICT = {}  # type: Dict[str, PluginEvent]

	@classmethod
	def register(cls, event: MCDREvent):
		cls.EVENT_DICT[event.id] = event
		return event

	@classmethod
	def get_event_list(cls) -> List[PluginEvent]:
		return list(cls.EVENT_DICT.values())


class MCDRPluginEvents:
	GENERAL_INFO 	= _PluginEventStorage.register(MCDREvent('mcdr.general_info', 'General info', 'on_info'))
	USER_INFO 		= _PluginEventStorage.register(MCDREvent('mcdr.user_info', 'User info', 'on_user_info'))
	SERVER_START 	= _PluginEventStorage.register(MCDREvent('mcdr.server_start', 'Server start', 'on_server_start'))
	SERVER_STARTUP 	= _PluginEventStorage.register(MCDREvent('mcdr.server_startup', 'Server startup', 'on_server_startup'))
	SERVER_STOP 	= _PluginEventStorage.register(MCDREvent('mcdr.server_stop', 'Server startup', 'on_server_stop'))
	MCDR_START 		= _PluginEventStorage.register(MCDREvent('mcdr.mcdr_start', 'MCDR start', 'on_mcdr_start'))
	MCDR_STOP 		= _PluginEventStorage.register(MCDREvent('mcdr.mcdr_stop', 'MCDR stop', 'on_mcdr_stop'))
	PLAYER_JOINED 	= _PluginEventStorage.register(MCDREvent('mcdr.player_joined', 'Player joined', 'on_player_joined'))
	PLAYER_LEFT 	= _PluginEventStorage.register(MCDREvent('mcdr.player_left', 'Player left', 'on_player_left'))
	PLUGIN_LOADED 	= _PluginEventStorage.register(MCDREvent('mcdr.plugin_loaded', 'Plugin loaded', 'on_load'))
	PLUGIN_UNLOADED = _PluginEventStorage.register(MCDREvent('mcdr.plugin_unloaded', 'Plugin unloaded', 'on_unload'))
	# PLUGIN_REMOVED 	= _PluginEventStorage.register(MCDREvent('mcdr.plugin_removed', 'Plugin removed', 'on_remove'))

	@classmethod
	def get_event_list(cls):
		return _PluginEventStorage.get_event_list()

	@classmethod
	def contains_id(cls, event_id: str):
		return event_id in _PluginEventStorage.EVENT_DICT


class EventListener:
	def __init__(self, plugin: 'AbstractPlugin', callback: Callable, priority: int):
		self.plugin = plugin
		self.callback = callback
		self.priority = priority

	def __lt__(self, other):
		if not isinstance(other, type(self)):
			return False
		return self.priority < other.priority

	def execute(self, *args, **kwargs):
		return self.callback(*args, **kwargs)

	def __repr__(self):
		return 'EventListener[plugin={},priority={},callback={}]'.format(self.plugin.get_name(), self.priority, self.callback)



import re
import time
from typing import Optional, Any, Union, Iterable, Tuple

from parse import parse

from mcdreforged.info_reactor.info import InfoSource, Info
from mcdreforged.info_reactor.server_information import ServerInformation
from mcdreforged.utils import string_util

'''
AbstractServerHandler
 ├─ BasicHandler
 ├─ AbstractMinecraftHandler
 │   ├─ VanillaHandler
 │   │   ├─ ForgeHandler
 │   │   └─ Beta18Handler
 │   └─ BukkitHandler
 │       ├─ Bukkit14Handler
 │       └─ CatServerHandler
 └─ BungeecordHandler
     └─ WaterfallHandler
'''


class AbstractServerHandler:
	# ---------------------
	#   Basic Information
	# ---------------------

	def get_name(self) -> str:
		return string_util.hump_to_underline(type(self).__name__)

	# ------------------
	#   Server control
	# ------------------

	def get_stop_command(self) -> str:
		"""
		The command to stop the server
		"""
		raise NotImplementedError()

	def get_send_message_command(self, target: str, message: Any, server_information: ServerInformation) -> Optional[str]:
		"""
		The command to send a message to a target
		"""
		raise NotImplementedError()

	def get_broadcast_message_command(self, message: Any, server_information: ServerInformation) -> Optional[str]:
		"""
		The command to broadcast a message in the server
		"""
		raise NotImplementedError()

	# -------------------------
	#   Server output parsing
	# -------------------------

	def pre_parse_server_stdout(self, text: str) -> str:
		"""
		Remove useless / annoying things like control characters in the text before parse

		:param str text: A line of the server stdout to be parsed
		:rtype: str
		"""
		return text

	@classmethod
	def parse_console_command(cls, text: str) -> Info:
		"""
		Base parsing, returns an almost un-parsed Info instance
		Don't use this unless

		:param str text: A line of the server stdout to be parsed
		:return: An Info instance
		:rtype: Info
		"""
		if type(text) is not str:
			raise TypeError('The text to parse should be a string')
		result = Info()
		result.raw_content = text
		t = time.localtime(time.time())
		result.hour = t.tm_hour
		result.min = t.tm_min
		result.sec = t.tm_sec
		result.content = text
		result.source = InfoSource.CONSOLE
		return result

	@classmethod
	def _get_server_stdout_raw_result(cls, text: str) -> Info:
		"""
		This is a raw parsing, returns an almost un-parsed Info instance
		Use as the first step of the parsing process, or as the return value if you give up parsing this text
		"""
		if type(text) is not str:
			raise TypeError('The text to parse should be a string')
		result = Info()
		result.source = InfoSource.SERVER
		result.raw_content = text
		result.content = string_util.clean_console_color_code(text)
		return result

	@classmethod
	def get_content_parsing_formatter(cls) -> Union[str, Iterable[str]]:
		"""
		Return a str or a str collection that is used in method _content_parse for parsing
		"""
		raise NotImplementedError()

	@classmethod
	def _content_parse(cls, info: Info):
		"""
		A common method to parse several elements from an un-parsed Info instance
		Elements expected to be parsed includes:
		- info.hour
		- info.min
		- info.sec
		- info.logging
		- info.content
		:param info: The incoming Info instance
		"""
		formatters = cls.get_content_parsing_formatter()
		if isinstance(formatters, str):
			formatters = (formatters,)
		for formatter in formatters:
			parsed = parse(formatter, info.content)
			if parsed is not None:
				logging_level = parsed['logging']
				if re.fullmatch(r'\w+', logging_level) is None:
					# logging level should be text only, just in case
					# might happens in e.g. WaterfallHandler parsing "[01:23:45 INFO] [Test]: ping"
					continue
				break
		else:
			raise ValueError('Unrecognized input: ' + info.content)
		info.hour = parsed['hour']
		info.min = parsed['min']
		info.sec = parsed['sec']
		info.logging_level = parsed['logging']
		info.content = parsed['content']

	def parse_server_stdout(self, text: str) -> Info:
		"""
		Main parsing operation. Parse a string from the stdout of the server
		It may raise any exceptions if the format of the input string is not correct

		In this implementation it achieves raw parsing, returns an almost un-parsed Info instance
		Use as the first step of the parsing process, or as the return value if you give up parsing this text
		:param str text: A line of the server stdout to be parsed
		:return: An Info instance
		:rtype: Info
		"""
		result = self._get_server_stdout_raw_result(text)
		self._content_parse(result)
		return result

	def parse_player_joined(self, info: Info) -> Optional[str]:
		"""
		Check if the info indicating a player joined message
		If it is, returns the name of the player, otherwise returns None

		:param Info info: The info instance that will be checked
		:return: The name of the player or None
		:rtype: str or None
		"""
		raise NotImplementedError()

	def parse_player_left(self, info: Info) -> Optional[str]:
		"""
		Check if the info indicates a player left message
		If it is, returns the name of the player, otherwise returns None

		:param Info info: The info instance that will be checked
		:return: The name of the player or None
		:rtype: str or None
		"""
		raise NotImplementedError()

	def parse_server_version(self, info: Info) -> Optional[str]:
		"""
		Check if the info contains a server version message
		If it is, returns server version, otherwise returns None

		:param Info info: The info instance that will be checked
		:return: The name of the player or None
		:rtype: str or None
		"""
		raise NotImplementedError()

	def parse_server_address(self, info: Info) -> Optional[Tuple[str, int]]:
		"""
		Check if the info contains the address which the server is listening on
		If it is, returns server ip and port, otherwise returns None

		:param Info info: The info instance that will be checked
		:return: A tuple containing the ip and the port
		"""
		raise NotImplementedError()

	def test_server_startup_done(self, info: Info) -> bool:
		"""
		Check if the info indicates a server startup message and return a bool

		:param Info info: The info instance that will be checked
		:return: If the info indicates a server startup message
		:rtype: bool
		"""
		raise NotImplementedError()

	def test_rcon_started(self, info: Info) -> bool:
		"""
		Check if rcon has started and return a bool

		:param Info info: The info instance that will be checked
		:return: If rcon has started
		:rtype: bool
		"""
		raise NotImplementedError()

	def test_server_stopping(self, info: Info) -> bool:
		"""
		Check if the server is stopping and return a bool

		:param Info info: The info instance that will be checked
		:return: If the server is stopping
		:rtype: bool
		"""
		raise NotImplementedError()



import os
import sys
import time
from abc import ABC
from typing import TYPE_CHECKING, Optional, List, Tuple, Any

from mcdreforged.plugin.meta.metadata import Metadata
from mcdreforged.plugin.plugin_event import MCDRPluginEvents, EventListener, PluginEvent
from mcdreforged.plugin.plugin_registry import DEFAULT_LISTENER_PRIORITY
from mcdreforged.plugin.type.plugin import AbstractPlugin, PluginState
from mcdreforged.utils.exception import IllegalCallError
from mcdreforged.utils.logger import DebugOption

if TYPE_CHECKING:
	from mcdreforged.plugin.plugin_manager import PluginManager

MODULE_TYPE = Any


class RegularPlugin(AbstractPlugin, ABC):
	def __init__(self, plugin_manager: 'PluginManager', file_path: str):
		super().__init__(plugin_manager, file_path)
		self.file_name = os.path.basename(file_path)
		self.file_modify_time = None  # type: Optional[int]
		self.__metadata = None  # type: Optional[Metadata]
		self.entry_module_instance = None  # type: MODULE_TYPE
		self.old_entry_module_instance = None  # type: MODULE_TYPE
		self.decorated_event_listeners = []  # type: List[Tuple[PluginEvent, EventListener]]

	def _reset(self):
		self.file_modify_time = self.calculate_file_modify_time()
		self.plugin_registry.clear()

	def is_regular(self) -> bool:
		return True

	def get_metadata(self) -> Metadata:
		if self.__metadata is None:
			raise IllegalCallError('Meta data of plugin {} is not loaded. Plugin state = {}'.format(repr(self), self.state))
		return self.__metadata

	def _set_metadata(self, metadata: Metadata):
		self.__metadata = metadata

	def get_fallback_metadata_id(self) -> str:
		raise NotImplementedError()

	@property
	def __class_name(self):
		return self.__class__.__name__

	def __repr__(self):
		return '{}[file={},path={},state={}]'.format(self.__class_name, self.file_name, self.plugin_path, self.state)

	# ----------------------
	#   Instance Operation
	# ----------------------

	def is_own_module(self, module_name: str) -> bool:
		raise NotImplementedError()

	def _load_entry_instance(self):
		self.old_entry_module_instance = self.entry_module_instance
		with self.plugin_manager.with_plugin_context(self):
			self.entry_module_instance = self._get_module_instance()

	# ---------------------
	#   To be Implemented
	# ---------------------

	def _get_module_instance(self):
		raise NotImplementedError()

	def _on_ready(self):
		self._register_default_listeners()

	def _on_load(self):
		self._reset()

	def _on_unload(self):
		for module_name in sys.modules.copy().keys():
			if self.is_own_module(module_name):
				rv = sys.modules.pop(module_name, None)
				self.mcdr_server.logger.debug('Removed module {} when unloading plugin {}, success = {}'.format(module_name, repr(self), rv is not None), option=DebugOption.PLUGIN)

	# --------------
	#   Life Cycle
	# --------------

	def load(self):
		self.assert_state({PluginState.UNINITIALIZED})
		self.set_state(PluginState.LOADING)
		self._on_load()
		self.mcdr_server.logger.debug('{} {} loaded from {}, file modify file = {}'.format(self.__class_name, self, self.plugin_path, self.pretty_file_modify_time), option=DebugOption.PLUGIN)
		self.set_state(PluginState.LOADED)

	def ready(self):
		"""
		Get ready, and register default things (listeners etc.)
		"""
		self.assert_state({PluginState.LOADED, PluginState.READY})
		self._on_ready()
		self.set_state(PluginState.READY)

	def reload(self):
		self.assert_state({PluginState.READY})
		self._on_unload()
		self._on_load()
		self.mcdr_server.logger.debug('{} {} reloaded, file modify file = {}'.format(self.__class_name, self, self.pretty_file_modify_time))

	def unload(self):
		self.assert_state({PluginState.LOADED, PluginState.READY})
		self._on_unload()
		self.set_state(PluginState.UNLOADING)

	def remove(self):
		self.assert_state({PluginState.UNLOADING})
		self.set_state(PluginState.UNLOADED)

	# -------------------
	#   Plugin Registry
	# -------------------

	def _register_default_listeners(self):
		for event in MCDRPluginEvents.get_event_list():
			if isinstance(event.default_method_name, str):
				func = getattr(self.entry_module_instance, event.default_method_name, None)
				if callable(func):
					self.register_event_listener(event, EventListener(self, func, DEFAULT_LISTENER_PRIORITY))
		for event, listener in self.decorated_event_listeners:
			self.register_event_listener(event, listener)
		self.decorated_event_listeners.clear()

	def register_event_listener(self, event: PluginEvent, listener: EventListener):
		# Special handling event listener registered with @event_listener decorator
		# Store and register them in method _register_default_listeners
		if self.in_states({PluginState.LOADING}):
			self.decorated_event_listeners.append((event, listener))
		else:
			super().register_event_listener(event, listener)

	# ---------------
	#   Plugin File
	# ---------------

	def plugin_exists(self):
		return os.path.isfile(self.plugin_path)

	def file_changed(self):
		return self.calculate_file_modify_time() != self.file_modify_time

	@property
	def pretty_file_modify_time(self) -> Optional[str]:
		if self.file_modify_time is not None:
			return time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(self.file_modify_time / 1e9))
		else:
			return None

	def calculate_file_modify_time(self):
		if self.plugin_exists():
			try:
				return os.stat(self.plugin_path).st_mtime_ns
			except:
				pass
		return None



"""
Core constants
"""

import os


NAME_SHORT = 'MCDR'
NAME = 'MCDReforged'
PACKAGE_NAME = 'mcdreforged'

# MCDR Version Storage
# Related: docs/source/conf.py
VERSION = '2.3.3'       # semver (1.2.3-alpha.4)
VERSION_PYPI = '2.3.3'  # pythonic ver (1.2.3a4)

GITHUB_URL = 'https://github.com/Fallen-Breath/MCDReforged'
GITHUB_API_LATEST = 'https://api.github.com/repos/Fallen-Breath/MCDReforged/releases/latest'
DOCUMENTION_URL = 'https://mcdreforged.readthedocs.io/'

PACKAGE_PATH = os.path.realpath(os.path.join(os.path.dirname(__file__), '..'))
LOGGING_FILE = os.path.join('logs', '{}.log'.format(NAME_SHORT))
LANGUAGE_FILE_SUFFIX = '.yml'
DEFAULT_LANGUAGE = 'en_us'

PLUGIN_THREAD_POOL_SIZE = 4
MAX_TASK_QUEUE_SIZE = 2048
WAIT_TIME_AFTER_SERVER_STDOUT_END_SEC = 60
REACTOR_QUEUE_FULL_WARN_INTERVAL_SEC = 5



import json
import re
from abc import ABC
from typing import Optional, Any

from parse import parse

from mcdreforged.handler.abstract_server_handler import AbstractServerHandler
from mcdreforged.info_reactor.info import Info
from mcdreforged.info_reactor.server_information import ServerInformation
from mcdreforged.minecraft.rtext import RTextBase
from mcdreforged.utils import string_util


class AbstractMinecraftHandler(AbstractServerHandler, ABC):
	def get_stop_command(self) -> str:
		return 'stop'

	@classmethod
	def format_message(cls, message: Any) -> str:
		if isinstance(message, RTextBase):
			return message.to_json_str()
		else:
			return json.dumps(str(message))

	def get_send_message_command(self, target: str, message: Any, server_information: ServerInformation) -> Optional[str]:
		can_do_execute = False
		try:
			from mcdreforged.plugin.meta.version import Version
			version = Version(server_information.version.split(' ')[0])
			if version >= Version('1.13.0'):
				can_do_execute = True
		except:
			pass
		command = 'tellraw {} {}'.format(target, self.format_message(message))
		if can_do_execute:
			command = 'execute at @p run ' + command
		return command

	def get_broadcast_message_command(self, message: Any, server_information: ServerInformation) -> Optional[str]:
		return self.get_send_message_command('@a', message, server_information)

	@classmethod
	def _get_server_stdout_raw_result(cls, text: str) -> Info:
		raw_result = super()._get_server_stdout_raw_result(text)
		# Minecraft <= 1.12.x might output minecraft color codes to the console
		# Just remove that
		raw_result.content = string_util.clean_minecraft_color_code(raw_result.content)
		return raw_result

	@classmethod
	def _verify_player_name(cls, name: str):
		return re.fullmatch(r'\w+', name) is not None

	def parse_server_stdout(self, text: str):
		result = super().parse_server_stdout(text)
		parsed = parse('<{name}> {message}', result.content)
		if parsed is not None and self._verify_player_name(parsed['name']):
			result.player, result.content = parsed['name'], parsed['message']
		return result

	def parse_player_joined(self, info: Info):
		# Steve[/127.0.0.1:9864] logged in with entity id 131 at (187.2703, 146.79014, 404.84718)
		if not info.is_user:
			parsed = parse('{name}[{}] logged in with entity id {} at ({})', info.content)
			if parsed is not None and self._verify_player_name(parsed['name']):
				return parsed['name']
		return None

	def parse_player_left(self, info: Info):
		# Steve left the game
		if not info.is_user and re.fullmatch(r'\w{1,16} left the game', info.content):
			return info.content.split(' ')[0]
		return None

	def parse_server_version(self, info: Info):
		if not info.is_user:
			parsed = parse('Starting minecraft server version {version}', info.content)
			if parsed is not None:
				return parsed['version']
		return None

	def parse_server_address(self, info: Info):
		if not info.is_user:
			parsed = parse('Starting Minecraft server on {}:{:d}', info.content)
			if parsed is not None:
				return parsed[0], parsed[1]
		return None

	def test_server_startup_done(self, info: Info):
		# 1.13+ Done (3.500s)! For help, type "help"
		# 1.13- Done (3.500s)! For help, type "help" or "?"
		return info.is_from_server and re.fullmatch(r'Done \([0-9.]*s\)! For help, type "help"( or "\?")?', info.content) is not None

	def test_rcon_started(self, info: Info):
		# RCON running on 0.0.0.0:25575
		return info.is_from_server and re.fullmatch(r'RCON running on [\w.]+:\d+', info.content) is not None

	def test_server_stopping(self, info: Info):
		# Stopping server
		return info.is_from_server and info.content == 'Stopping server'



"""
Info and InfoSource
"""
from typing import TYPE_CHECKING, Optional

from mcdreforged.command.command_source import ConsoleCommandSource, PlayerCommandSource, \
	InfoCommandSource
from mcdreforged.utils.exception import IllegalStateError, IllegalCallError

if TYPE_CHECKING:
	from mcdreforged.mcdr_server import MCDReforgedServer
	from mcdreforged.plugin.server_interface import ServerInterface


class InfoSource:
	# the text is from the stdout of the server
	SERVER = 0

	# the text is from user input
	CONSOLE = 1


class Info:
	__id_counter = 0

	def __init__(self):
		# a increasing id number for distinguishing info instance
		self.id = Info.__id_counter
		Info.__id_counter += 1
		self.__mcdr_server = None  # type: Optional[MCDReforgedServer]
		self.__send_to_server = True
		self.__command_source = None

		# -----------------
		#   Public fields
		# -----------------

		# time information from the parsed text
		self.hour = None  # type: Optional[int]
		self.min = None  # type: Optional[int]
		self.sec = None  # type: Optional[int]

		# very raw content, it's also the content to be echoed to the stdout
		self.raw_content = None  # type: Optional[str]

		# if the text is sent by a player the value will be what the player said. if not the value will be the pain text
		self.content = None  # type: Optional[str]

		# the name of the player. if it's not sent by a player the value will be None
		self.player = None  # type: Optional[str]

		# the value type is InfoSource
		self.source = None  # type: Optional[int]

		# the logging level of the server's stdout, such as "INFO" or "WARN"
		self.logging_level = None  # type: Optional[str]

	@property
	def is_from_console(self):
		return self.source == InfoSource.CONSOLE

	@property
	def is_from_server(self):
		return self.source == InfoSource.SERVER

	@property
	def is_player(self):
		return self.is_from_server and self.player is not None

	@property
	def is_user(self):
		return self.is_from_console or self.is_player

	# --------------
	#      API
	# --------------

	def __assert_attached(self):
		if self.__mcdr_server is None:
			raise IllegalStateError('MCDR server is not attached to this Info instance yet')

	def attach_mcdr_server(self, mcdr_server):
		if self.__mcdr_server is not None:
			raise IllegalStateError('An Info instance can only attach the MCDR server once')
		self.__mcdr_server = mcdr_server

	def get_server(self) -> 'ServerInterface':
		return self.__mcdr_server.basic_server_interface

	def get_command_source(self) -> Optional[InfoCommandSource]:
		self.__assert_attached()
		if self.__command_source is None:
			if self.is_from_console:
				self.__command_source = ConsoleCommandSource(self.__mcdr_server, self)
			elif self.is_player:
				self.__command_source = PlayerCommandSource(self.__mcdr_server, self, self.player)
		return self.__command_source

	def to_command_source(self) -> InfoCommandSource:
		source = self.get_command_source()
		if source is None:
			raise IllegalCallError()
		return source

	def should_send_to_server(self) -> bool:
		return self.__send_to_server

	def cancel_send_to_server(self) -> None:
		self.__send_to_server = False

	# --------------------------------
	#   Formatting and Magic methods
	# --------------------------------

	def format_text(self):
		try:
			time_message = '{:0>2}:{:0>2}:{:0>2}'.format(self.hour, self.min, self.sec)
		except:
			time_message = 'Invalid'
		return '\n'.join([
			'Time: {}; ID: {}'.format(time_message, self.id),
			'Player: {}; Source: {}; Logging level: {}'.format(self.player, self.source, self.logging_level),
			'Content: {}'.format(self.content),
			'Raw content: {}'.format(self.raw_content)
		])

	def __str__(self):
		return '; '.join(self.format_text().splitlines())

	def __deepcopy__(self, memo):
		"""
		Just dont copy the mcdr_server instance
		"""
		existed = memo.get(self)
		if existed is not None:
			return existed
		memo[self] = dupe = Info()
		dupe.hour, dupe.min, dupe.sec = self.hour, self.min, self.sec
		dupe.raw_content = self.raw_content
		dupe.content = self.content
		dupe.player = self.player
		dupe.source = self.source
		dupe.logging_level = self.logging_level
		return dupe



import re
from typing import Optional, Union, Iterable, Any

from parse import parse

from mcdreforged.handler.abstract_server_handler import AbstractServerHandler
from mcdreforged.info_reactor.info import Info
from mcdreforged.info_reactor.server_information import ServerInformation


class VelocityHandler(AbstractServerHandler):
	def get_stop_command(self) -> str:
		return 'shutdown'

	def get_send_message_command(self, target: str, message: Any, server_information: ServerInformation) -> Optional[str]:
		return None

	def get_broadcast_message_command(self, message: Any, server_information: ServerInformation) -> Optional[str]:
		return None

	@classmethod
	def get_content_parsing_formatter(cls) -> Union[str, Iterable[str]]:
		# It's the same as WaterfallHandler, but since it's a different server here comes an individual copy
		return (
			'[{hour:d}:{min:d}:{sec:d} {logging}]: {content}',
			'[{hour:d}:{min:d}:{sec:d} {logging}] {dummy}: {content}'  # somethings there is an extra element after the heading [] and :
		)

	def parse_player_joined(self, info: Info) -> Optional[str]:
		# [connected player] Fallen_Breath (/127.0.0.1:12896) has connected
		if not info.is_user:
			parsed = parse('[connected player] {name} (/{address}) has connected', info.content)
			if parsed is not None:
				return parsed['name']
		return None

	def parse_player_left(self, info: Info) -> Optional[str]:
		# [connected player] Fallen_Breath (/127.0.0.1:12896) has disconnected
		if not info.is_user:
			parsed = parse('[connected player] {name} (/{address}) has disconnected', info.content)
			if parsed is not None:
				return parsed['name']
		return None

	def parse_server_version(self, info: Info):
		return None

	def parse_server_address(self, info: Info):
		# Listening on /192.168.0.1:25577
		# Listening on /[0:0:0:0:0:0:0:0%0]:25577
		# Listening on /0:0:0:0:0:0:0:0%0:25577
		if not info.is_user:
			parsed = parse('Listening on /{}:{:d}', info.content)
			if parsed is not None:
				return parsed[0], parsed[1]
		return None

	def test_server_startup_done(self, info: Info) -> bool:
		# Done (3.05s)!
		return not info.is_user and re.fullmatch(r'Done \([0-9.]*s\)!', info.content) is not None

	def test_rcon_started(self, info: Info) -> bool:
		return False

	def test_server_stopping(self, info: Info) -> bool:
		# Shutting down the proxy...
		return not info.is_user and info.content == 'Shutting down the proxy...'



# Link to mcdreforged.rtext
# The advance text component class for Minecraft
from mcdreforged.minecraft.rtext import RColor, RTextBase, RStyle, RText, RAction, RTextTranslation, RTextList
from mcdreforged.translation.translation_text import RTextMCDRTranslation

__all__ = [
	'RColor', 'RStyle', 'RAction',
	'RTextBase', 'RText', 'RTextTranslation', 'RTextList',
	'RTextMCDRTranslation'
]



import traceback
from abc import ABC
from typing import TYPE_CHECKING, NamedTuple, Any, Callable

from mcdreforged.command.builder.exception import CommandError
from mcdreforged.command.builder.nodes.basic import Literal
from mcdreforged.command.command_source import CommandSource
from mcdreforged.permission.permission_level import PermissionLevel
from mcdreforged.translation.translation_text import RTextMCDRTranslation

if TYPE_CHECKING:
	from mcdreforged.mcdr_server import MCDReforgedServer
	from mcdreforged.plugin.builtin.mcdreforged_plugin.mcdreforged_plugin import MCDReforgedPlugin
	from mcdreforged.plugin.server_interface import PluginServerInterface


class FunctionCallResult(NamedTuple):
	return_value: Any
	no_error: bool


class SubCommand(ABC):
	def __init__(self, mcdr_plugin: 'MCDReforgedPlugin'):
		self.mcdr_plugin = mcdr_plugin

	def get_command_node(self) -> Literal:
		raise NotImplementedError()

	# ------------------
	#    Common Utils
	# ------------------

	def tr(self, key: str, *args, **kwargs) -> RTextMCDRTranslation:
		return self.mcdr_plugin.tr(key, *args, **kwargs)

	@property
	def mcdr_server(self) -> 'MCDReforgedServer':
		return self.mcdr_plugin.mcdr_server

	@property
	def server_interface(self) -> 'PluginServerInterface':
		return self.mcdr_plugin.server_interface

	@property
	def control_command_prefix(self):
		return self.mcdr_plugin.control_command_prefix

	@property
	def help_command_prefix(self):
		return self.mcdr_plugin.help_command_prefix

	@staticmethod
	def control_command_root(literal):
		return Literal(literal).requires(lambda src: src.has_permission(PermissionLevel.MCDR_CONTROL_LEVEL))

	@staticmethod
	def public_command_root(literal):
		return Literal(literal).requires(lambda src: src.has_permission(PermissionLevel.USER))

	@staticmethod
	def can_see_rtext(source: CommandSource):
		return source.is_player

	# ---------------------------------
	#     Refs to MCDReforgedPlugin
	# ---------------------------------

	def get_help_message(self, source: CommandSource, translation_key: str):
		return self.mcdr_plugin.get_help_message(source, translation_key)

	def on_mcdr_command_permission_denied(self, source: CommandSource, error: CommandError):
		self.mcdr_plugin.on_mcdr_command_permission_denied(source, error)

	def on_mcdr_command_unknown_argument(self, source: CommandSource, error: CommandError):
		self.mcdr_plugin.on_mcdr_command_unknown_argument(source, error)

	def _print_plugin_operation_result_if_no_error(self, source: CommandSource, ret: FunctionCallResult):
		if ret.no_error:
			source.reply(self.mcdr_server.plugin_manager.last_operation_result.to_rtext(show_path=source.has_permission(PermissionLevel.PHYSICAL_SERVER_CONTROL_LEVEL)))

	def function_call(self, source: CommandSource, func: Callable[[], Any], name: str, log_success=True, log_fail=True, msg_args=()) -> FunctionCallResult:
		try:
			ret = FunctionCallResult(func(), True)
			if log_success:
				source.reply(self.tr('mcdr_command.{}.success'.format(name), *msg_args))
			return ret
		except:
			if log_fail:
				source.reply(self.tr('mcdr_command.{}.fail'.format(name), *msg_args))
			self.mcdr_server.logger.error(traceback.format_exc())
			return FunctionCallResult(None, False)



"""
MCDR update things
"""
import time
from threading import Lock
from typing import Callable, Any, Union, Optional

import requests

from mcdreforged.constants import core_constant
from mcdreforged.executor.thread_executor import ThreadExecutor
from mcdreforged.minecraft.rtext import RText, RAction, RColor, RStyle, RTextBase
from mcdreforged.plugin.meta.version import Version
from mcdreforged.translation.translation_text import RTextMCDRTranslation
from mcdreforged.utils import misc_util


class UpdateHelper(ThreadExecutor):
	def __init__(self, mcdr_server):
		super().__init__(mcdr_server)
		self.__api_fetcher = GithubApiFetcher(core_constant.GITHUB_API_LATEST)
		self.__update_lock = Lock()
		self.__last_query_time = 0

	def tick(self):
		if time.monotonic() - self.__last_query_time >= 60 * 60 * 24:
			self.check_update(lambda: self.mcdr_server.config['check_update'] is True, self.mcdr_server.logger.info)
		time.sleep(1)

	def check_update(self, condition_check: Callable[[], bool], reply_func: Callable[[Union[str or RTextBase]], Any]):
		self.__last_query_time = time.monotonic()
		misc_util.start_thread(self.__check_update, (condition_check, reply_func), 'CheckUpdate')

	def tr(self, key: str, *args, **kwargs) -> RTextBase:
		return RTextMCDRTranslation(key, *args, **kwargs).set_translator(self.mcdr_server.tr)

	def __check_update(self, condition_check: Callable[[], bool], reply_func: Callable[[Union[str or RTextBase]], Any]):
		if not condition_check():
			return
		acquired = self.__update_lock.acquire(blocking=False)
		if not acquired:
			reply_func(self.tr('update_helper.check_update.already_checking'))
			return
		try:
			response = None
			try:
				response = self.__api_fetcher.fetch()
				latest_version: str = response['tag_name']
				update_log: str = response['body']
			except Exception as e:
				reply_func(self.tr('update_helper.check_update.check_fail', repr(e)))
				if isinstance(e, KeyError) and isinstance(response, dict) and 'message' in response:
					reply_func(response['message'])
					if 'documentation_url' in response:
						reply_func(
							RText(response['documentation_url'], color=RColor.blue, styles=RStyle.underlined).
							h(response['documentation_url']).
							c(RAction.open_url, response['documentation_url'])
						)
			else:
				try:
					version_current = Version(core_constant.VERSION, allow_wildcard=False)
					version_fetched = Version(latest_version.lstrip('vV'), allow_wildcard=False)
				except:
					self.mcdr_server.logger.exception('Fail to compare between versions "{}" and "{}"'.format(core_constant.VERSION, latest_version))
					return
				if version_current == version_fetched:
					reply_func(self.tr('update_helper.check_update.is_already_latest'))
				elif version_current > version_fetched:
					reply_func(self.tr('update_helper.check_update.newer_than_latest', core_constant.VERSION, latest_version))
				else:
					reply_func(self.tr('update_helper.check_update.new_version_detected', latest_version))
					for line in update_log.splitlines():
						reply_func('    {}'.format(line))
		finally:
			self.__update_lock.release()


class GithubApiFetcher:
	def __init__(self, url: str):
		self.url = url
		self.__etag = 'dummy'
		self.__cached_response: Optional[dict] = None

	def fetch(self) -> dict:
		response = requests.get(self.url, headers={'If-None-Match': self.__etag}, timeout=10)
		self.__etag = response.headers.get('ETag', self.__etag)
		if response.status_code != 304:  # 304 means content keeps unchanged
			if response.status_code != 200:
				raise Exception('Un-expected status code {}: {}'.format(response.status_code, response.content))
			self.__cached_response = response.json()
		return self.__cached_response
